# This file is part of cloud-init. See LICENSE file for license information.
# Automatically generated by Meson. Do not edit this file directly

UPSTREAM_VERSION = "26.2"      # Required to release
DOWNSTREAM_VERSION = "@DOWNSTREAM_VERSION@"  # Optional for downstream packagers
